from __future__ import annotations

import inspect
import json
import logging
import time
from enum import Enum
from typing import Set

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from victron_ble.devices import Device, DeviceData, detect_device_type

logger = logging.getLogger(__name__)


class BaseScanner:
    def __init__(self) -> None:
        """Initialize the scanner."""
        self._scanner: BleakScanner = BleakScanner(
            detection_callback=self._detection_callback
        )
        self._seen_data: Set[bytes] = set()

    def _detection_callback(self, device: BLEDevice, advertisement: AdvertisementData):
        # Filter for Victron devices and instant readout advertisements
        data = advertisement.manufacturer_data.get(0x02E1)
        if not data or not data.startswith(b"\x10") or data in self._seen_data:
            return

        # De-duplicate advertisements
        if len(self._seen_data) > 1000:
            self._seen_data = set()
        self._seen_data.add(data)

        self.callback(device, data)

    def callback(self, device: BLEDevice, data: bytes):
        raise NotImplementedError()

    async def start(self):
        await self._scanner.start()

    async def stop(self):
        await self._scanner.stop()


# An ugly hack to print a class as JSON
class DeviceDataEncoder(json.JSONEncoder):
    def default(self, obj):
        if issubclass(obj.__class__, DeviceData):
            data = {}
            for name, method in inspect.getmembers(obj, predicate=inspect.ismethod):
                if name.startswith("get_"):
                    value = method()
                    if isinstance(value, Enum):
                        value = value.name.lower()
                    if value is not None:
                        data[name[4:]] = value
            return data


class Scanner(BaseScanner):
    def __init__(self, device_keys: dict[str, str] = {}):
        super().__init__()
        self._device_keys = {k.lower(): v for k, v in device_keys.items()}
        self._known_devices: dict[str, Device] = {}

    async def start(self):
        logger.info(f"Reading data for {list(self._device_keys.keys())}")
        await super().start()

    def callback(self, device: BLEDevice, data: bytes):
        advertisement_key = self._device_keys.get(device.address.lower())
        if advertisement_key is None:
            return

        if device.address not in self._known_devices:
            device_klass = detect_device_type(data)
            if not device_klass:
                logger.error(f"Could not identify device type for {device}")
                return

            self._known_devices[device.address] = device_klass(advertisement_key)

        logger.debug(f"Received data from {device.address.lower()}: {data.hex()}")
        parsed = self._known_devices[device.address].parse(data)
        blob = {
            "name": device.name,
            "address": device.address,
            "rssi": device.rssi,
            "payload": parsed,
        }
        print(json.dumps(blob, cls=DeviceDataEncoder, indent=2))


class DiscoveryScanner(BaseScanner):
    def __init__(self) -> None:
        super().__init__()
        self._seen_devices: Set[str] = set()

    def callback(self, device: BLEDevice, advertisement: bytes):
        if device.address not in self._seen_devices:
            logger.info(f"{device}")
            self._seen_devices.add(device.address)


class DebugScanner(BaseScanner):
    def __init__(self, address: str):
        super().__init__()
        self.address = address

    async def start(self):
        logger.info(f"Dumping advertisements from {self.address}")
        await super().start()

    def callback(self, device: BLEDevice, data: bytes):
        if device.address.lower() == self.address.lower():
            logger.info(f"{time.time():<24}: {data.hex()}")



from victron_ble.devices.battery_monitor import BatteryMonitor, AuxMode


class TestBatteryMonitor:
    def test_parse_data(self) -> None:
        data = "100289a302b040af925d09a4d89aa0128bdef48c6298a9"
        actual = BatteryMonitor("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )
        assert actual.get_aux_mode() == AuxMode.DISABLED
        assert actual.get_consumed_ah() == 50.0
        assert actual.get_current() == 0
        assert actual.get_remaining_mins() == 65535
        assert actual.get_soc() == 50.0
        assert actual.get_voltage() == 12.53
        assert actual.get_high_starter_battery_voltage_alarm() == False
        assert actual.get_high_temperature_alarm() == False
        assert actual.get_high_voltage_alarm() == False
        assert actual.get_low_soc_alarm() == False
        assert actual.get_low_starter_battery_voltage_alarm() == False
        assert actual.get_low_temperature_alarm() == False
        assert actual.get_low_voltage_alarm() == False
        assert actual.get_midpoint_deviation_alarm() == False

        assert actual.get_temperature() == None
        assert actual.get_starter_voltage() == None
        assert actual.get_midpoint_voltage() == None

    def test_aux_midpoint(self) -> None:
        data = "100289a3021001afc15f433b2663c8cfc0678b5d3d29a8"
        actual = BatteryMonitor("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )
        assert actual.get_midpoint_voltage() == 655.34

    def test_aux_starter(self) -> None:
        data = "100289a302c802af45fc59d010dd78d2948e0c55c3bf48"
        actual = BatteryMonitor("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )
        assert actual.get_starter_voltage() == -0.02

    def test_aux_temperature(self) -> None:
        data = "100289a302bb01af129087600b9b97bc2c32867c8238da"
        actual = BatteryMonitor("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )
        assert actual.get_temperature() == 382.2



from enum import Enum
from typing import Any, Dict, Optional

from construct import (
    BitStruct,
    Flag,
    GreedyBytes,
    Int8sl,
    Int16sl,
    Int16ul,
    Int24sl,
    Struct,
)

from victron_ble.devices.base import Device, DeviceData, kelvin_to_celsius


class AuxMode(Enum):
    STARTER_VOLTAGE = 0
    MIDPOINT_VOLTAGE = 1
    TEMPERATURE = 2
    DISABLED = 3


class BatteryMonitorData(DeviceData):
    def __init__(self, data: Dict[str, Any]) -> None:
        self._data = data

    def get_remaining_mins(self) -> float:
        """
        Return the number of remaining minutes of battery life in minutes
        """
        return self._data["remaining_mins"]

    def get_current(self) -> float:
        """
        Return the current in amps
        """
        return self._data["current"]

    def get_voltage(self) -> float:
        """
        Return the voltage in volts
        """
        return self._data["voltage"]

    def get_soc(self) -> float:
        """
        Return the state of charge in percentage
        """
        return self._data["soc"]

    def get_consumed_ah(self) -> float:
        """
        Return the consumed energy in amp hours
        """
        return self._data["consumed_ah"]

    def get_low_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the low voltage alarm is active
        """
        return self._data["alarm"]["low_voltage"]

    def get_high_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the high voltage alarm is active
        """
        return self._data["alarm"]["high_voltage"]

    def get_low_soc_alarm(self) -> bool:
        """
        Return a boolean indicating if the low state of charge alarm is active
        """
        return self._data["alarm"]["low_soc"]

    def get_low_starter_battery_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the low starter battery voltage alarm is active
        """
        return self._data["alarm"]["low_starter_voltage"]

    def get_high_starter_battery_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the high starter battery voltage alarm is active
        """
        return self._data["alarm"]["high_starter_voltage"]

    def get_low_temperature_alarm(self) -> bool:
        """
        Return a boolean indicating if the low temperature alarm is active
        """
        return self._data["alarm"]["low_temperature"]

    def get_high_temperature_alarm(self) -> bool:
        """
        Return a boolean indicating if the high temperature alarm is active
        """
        return self._data["alarm"]["high_temperature"]

    def get_midpoint_deviation_alarm(self) -> bool:
        """
        Return a boolean indicating if the high temperature alarm is active
        """
        return self._data["alarm"]["mid_deviation"]

    def get_aux_mode(self) -> AuxMode:
        """
        Return an enum indicating the current auxiliary input mode
        """
        return self._data["aux_mode"]

    def get_temperature(self) -> Optional[float]:
        """
        Return the temperature in Celsius if the aux input is set to temperature
        """
        temp = self._data.get("temperature_kelvin")
        if temp:
            return kelvin_to_celsius(temp)
        return None

    def get_starter_voltage(self) -> Optional[float]:
        """
        Return the starter battery voltage in volts if the aux input is set to starter battery
        """
        return self._data.get("starter_voltage")

    def get_midpoint_voltage(self) -> Optional[float]:
        """
        Return the midpoint battery voltage in volts if the aux input is set to midpoint voltage
        """
        return self._data.get("midpoint_voltage")


class BatteryMonitor(Device):

    PACKET = Struct(
        # Remaining time in minutes
        "remaining_mins" / Int16ul,
        # Voltage reading in 0.1v increments
        "voltage" / Int16ul,
        # Bit map of alarm status
        # 0b00000001 = low voltage alarm
        # 0b00000010 = high voltage alarm
        # 0b00000100 = low soc alarm
        # 0b00001000 = low starter alarm
        # 0b00001001 = low voltage + low starter alarm
        # 0b00011001 = high starter + low voltage + low starter alarm
        # 0b00010000 = high starter alarm
        # 0b00100000 = low temp alarm
        # 0b01000000 = high temp alarm
        # 0b10000000 = midpoint voltage deviation alarm
        "alarm"
        / BitStruct(
            "mid_deviation" / Flag,
            "high_temperature" / Flag,
            "low_temperature" / Flag,
            "high_starter_voltage" / Flag,
            "low_starter_voltage" / Flag,
            "low_soc" / Flag,
            "high_voltage" / Flag,
            "low_voltage" / Flag,
        ),
        # Unknown byte
        "uk_1b" / Int8sl,
        # Value of the auxillary input (millivolts or degrees)
        "aux" / Int16ul,
        # The upper 22 bits indicate the current in milliamps
        # The lower 2 bits identify the aux input mode:
        #   0 = Starter battery voltage
        #   1 = Midpoint voltage
        #   2 = Temperature
        #   3 = Disabled
        "current" / Int24sl,
        # Consumed Ah in 0.1Ah increments
        "consumed_ah" / Int16ul,
        # The lowest 4 bits are unknown
        # The next 8 bits indicate the state of charge in 0.1% increments
        # The upper 2 bits are unknown
        "soc" / Int16ul,
        # Throw away any extra bytes
        GreedyBytes,
    )

    def parse(self, data: bytes) -> BatteryMonitorData:
        decrypted = self.decrypt(data)
        pkt = self.PACKET.parse(decrypted)

        aux_mode = AuxMode(pkt.current & 0b11)

        parsed = {
            "remaining_mins": pkt.remaining_mins,
            "aux_mode": aux_mode,
            "current": (pkt.current >> 2) / 1000,
            "voltage": pkt.voltage / 100,
            "consumed_ah": pkt.consumed_ah / 10,
            "soc": ((pkt.soc & 0x3FFF) >> 4) / 10,
            "alarm": {
                "low_voltage": pkt.alarm.low_voltage,
                "high_voltage": pkt.alarm.high_voltage,
                "low_soc": pkt.alarm.low_soc,
                "low_starter_voltage": pkt.alarm.low_starter_voltage,
                "high_starter_voltage": pkt.alarm.high_starter_voltage,
                "low_temperature": pkt.alarm.low_temperature,
                "high_temperature": pkt.alarm.high_temperature,
                "mid_deviation": pkt.alarm.mid_deviation,
            },
        }

        if aux_mode == AuxMode.STARTER_VOLTAGE:
            # Starter voltage is treated as signed
            parsed["starter_voltage"] = (
                Int16sl.parse((pkt.aux).to_bytes(2, "little")) / 100
            )
        elif aux_mode == AuxMode.MIDPOINT_VOLTAGE:
            parsed["midpoint_voltage"] = pkt.aux / 100
        elif aux_mode == AuxMode.TEMPERATURE:
            parsed["temperature_kelvin"] = pkt.aux / 100

        return BatteryMonitorData(parsed)



from victron_ble.devices import (
    BatteryMonitor,
    BatterySense,
    DcEnergyMeter,
    detect_device_type,
)


def test_unknown_device_type() -> None:
    assert detect_device_type(bytes.fromhex("1002000000")) == None


def test_battery_monitor_discovery() -> None:
    assert detect_device_type(bytes.fromhex("100289a302")) == BatteryMonitor


def test_dc_energy_discovery() -> None:
    assert detect_device_type(bytes.fromhex("100289a30d")) == DcEnergyMeter


def test_battery_sense_discovery() -> None:
    assert detect_device_type(bytes.fromhex("1002a4a302")) == BatterySense



from victron_ble.devices import BatterySense


class TestBatterySense:
    def test_parse_data(self) -> None:
        key = "0da694539597f9cf6c613cde60d7bf05"
        data = "1000a4a3025f150d8dcbff517f30eb65e76b22a04ac4e1"
        actual = BatterySense(key).parse(bytes.fromhex(data))
        assert actual.get_temperature() == 22.5
        assert actual.get_voltage() == 12.22



import abc

from construct import FixedSized, GreedyBytes, Int8sl, Int16ul, Struct
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad


class DeviceData:
    pass


class Device(abc.ABC):
    def __init__(self, advertisement_key: str):
        self.advertisement_key = advertisement_key

    def decrypt(self, data: bytes) -> bytes:
        parser = Struct(
            "prefix" / FixedSized(2, GreedyBytes),
            # Model ID
            "device_id" / Int16ul,
            # Packet type
            "readout_type" / Int8sl,
            # IV for encryption
            "iv" / Int16ul,
            "encrypted_data" / GreedyBytes,
        )

        container = parser.parse(data)

        # The first byte of advertised data seems to match the first byte of the advertisement key
        cipher = AES.new(
            bytes.fromhex(self.advertisement_key),
            AES.MODE_OFB,
            iv=container.iv.to_bytes(16, "little"),
        )
        return cipher.decrypt(pad(container.encrypted_data[1:], 16))

    @abc.abstractmethod
    def parse(self, data: bytes):
        pass


def kelvin_to_celsius(temp_in_kelvin: float) -> float:
    return round(temp_in_kelvin - 273.15, 2)



from typing import Any, Dict

from victron_ble.devices import BatteryMonitor, Device, DeviceData


class BatterySenseData(DeviceData):
    def __init__(self, data: Dict[str, Any]) -> None:
        self._data = data

    def get_temperature(self) -> float:
        """
        Return the temperature in Celsius
        """
        return self._data["temperature"]

    def get_voltage(self) -> float:
        """
        Return the voltage in volts
        """
        return self._data["voltage"]


class BatterySense(Device):
    def parse(self, data: bytes) -> BatterySenseData:
        parsed = BatteryMonitor(self.advertisement_key).parse(data)

        return BatterySenseData(
            {"temperature": parsed.get_temperature(), "voltage": parsed.get_voltage()}
        )



from victron_ble.devices.battery_monitor import AuxMode
from victron_ble.devices.dc_energy_meter import DcEnergyMeter, MeterType


class TestDcEnergyMeter:
    def test_parse_data(self) -> None:
        data = "100289a30d787fafde83ccec982199fd815286"
        actual = DcEnergyMeter("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )

        assert actual.get_meter_type() == MeterType.FUEL_CELL
        assert actual.get_aux_mode() == AuxMode.STARTER_VOLTAGE
        assert actual.get_current() == 0.0
        assert actual.get_voltage() == 12.52
        assert actual.get_starter_voltage() == -0.01

        assert actual.get_high_starter_battery_voltage_alarm() == False
        assert actual.get_low_starter_battery_voltage_alarm() == False
        assert actual.get_high_voltage_alarm() == False
        assert actual.get_low_voltage_alarm() == False
        assert actual.get_high_temperature_alarm() == False
        assert actual.get_low_temperature_alarm() == False

        assert actual.get_temperature() == None

    def test_aux_starter(self) -> None:
        data = "100289a30d787fafde83ccec982199fd815286"
        actual = DcEnergyMeter("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )
        assert actual.get_starter_voltage() == -0.01

    def test_aux_temperature(self) -> None:
        data = "108289a30df07faf9629bfb8c0153f431362c4"
        actual = DcEnergyMeter("aff4d0995b7d1e176c0c33ecb9e70dcd").parse(
            bytes.fromhex(data)
        )
        assert actual.get_temperature() == 382.2



from typing import Dict, Optional, Type

from construct import Int8ul, Int16ul

from victron_ble.devices.base import Device, DeviceData
from victron_ble.devices.battery_monitor import AuxMode, BatteryMonitor
from victron_ble.devices.battery_sense import BatterySense
from victron_ble.devices.dc_energy_meter import DcEnergyMeter

__all__ = ["AuxMode", "Device", "DeviceData", "BatteryMonitor", "DcEnergyMeter"]

MODEL_MAPPING: Dict[int, Type[Device]] = {
    0xA3A4: BatterySense,  # Smart Battery Sense
}


def detect_device_type(data: bytes) -> Optional[Type[Device]]:
    model_id = Int16ul.parse(data[2:4])
    mode = Int8ul.parse(data[4:5])

    # Model ID-based preferences
    match = MODEL_MAPPING.get(model_id)
    if match:
        return match

    # Defaults
    if mode == 0x2:  # BatteryMonitor
        return BatteryMonitor
    elif mode == 0xD:  # DcEnergyMeter
        return DcEnergyMeter
    elif mode == 0x8:  # AcCharger
        pass
    elif mode == 0x4:  # DcDcConverter
        pass
    elif mode == 0x3:  # Inverter
        pass
    elif mode == 0x6:  # InverterRS
        pass
    elif mode == 0xA:  # LynxSmartBMS
        pass
    elif mode == 0xB:  # MultiRS
        pass
    elif mode == 0x5:  # SmartLithium
        pass
    elif mode == 0x1:  # SolarCharger
        pass
    elif mode == 0xC:  # VE.Bus
        pass

    return None



import asyncio
import logging
from typing import List, Tuple

import click

from victron_ble.scanner import DebugScanner, DiscoveryScanner, Scanner

logger = logging.getLogger("victron_ble")
logging.basicConfig()


class DeviceKeyParam(click.ParamType):
    name = "device_key"

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            parts = value.split(":")
            if len(parts) == 2:
                addr = parts[0].strip()
                key = parts[1].strip()
                return (addr, key)

        self.fail(f"{value} is not a valid <addr>:<key> pair", param, ctx)


@click.group()
@click.option("-v", "--verbose", is_flag=True, help="Increase logging output")
def cli(verbose):
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)


@cli.command(help="Discover Victron devices with Instant Readout")
def discover():
    loop = asyncio.get_event_loop()

    async def scan():
        scanner = DiscoveryScanner()
        await scanner.start()

    asyncio.ensure_future(scan())
    loop.run_forever()


@cli.command(help="Dump all advertisements matching the given device ID")
@click.argument("id", type=str)
def dump(id: str):
    loop = asyncio.get_event_loop()

    async def scan():
        scanner = DebugScanner(id)
        await scanner.start()

    asyncio.ensure_future(scan())
    loop.run_forever()


@cli.command(help="Read data from specified devices")
@click.argument("device_keys", nargs=-1, type=DeviceKeyParam())
def read(device_keys: List[Tuple[str, str]]):
    loop = asyncio.get_event_loop()

    async def scan(keys):
        scanner = Scanner(keys)
        await scanner.start()

    asyncio.ensure_future(scan({k: v for k, v in device_keys}))
    loop.run_forever()


if __name__ == "__main__":
    cli()



from enum import Enum
from typing import Any, Dict, Optional

from construct import BitStruct, Flag, Int8sl, Int16sl, Int16ul, Int24sl, Struct

from victron_ble.devices.base import Device, DeviceData, kelvin_to_celsius
from victron_ble.devices.battery_monitor import AuxMode


class MeterType(Enum):
    SOLAR_CHARGER = -7
    WIND_CHARGER = -6
    SHAFT_GENERATOR = -5
    ALTERNATOR = -4
    FUEL_CELL = -3
    WATER_GENERATOR = -2
    DC_DC_CHARGER = -1
    AC_CHARGER = 1
    GENERIC_SOURCE = 2
    GENERIC_LOAD = 3
    ELECTRIC_DRIVE = 4
    FRIDGE = 5
    WATER_PUMP = 6
    BILGE_PUMP = 7
    DC_SYSTEM = 8
    INVERTER = 9
    WATER_HEATER = 10


class DcEnergyMeterData(DeviceData):
    def __init__(self, data: Dict[str, Any]) -> None:
        self._data = data

    def get_meter_type(self) -> MeterType:
        """
        Return an enum indicating the current meter type
        """
        return self._data["meter_type"]

    def get_current(self) -> float:
        """
        Return the current in amps
        """
        return self._data["current"]

    def get_voltage(self) -> float:
        """
        Return the voltage in volts
        """
        return self._data["voltage"]

    def get_low_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the low voltage alarm is active
        """
        return self._data["alarm"]["low_voltage"]

    def get_high_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the high voltage alarm is active
        """
        return self._data["alarm"]["high_voltage"]

    def get_low_starter_battery_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the low starter battery voltage alarm is active
        """
        return self._data["alarm"]["low_starter_voltage"]

    def get_high_starter_battery_voltage_alarm(self) -> bool:
        """
        Return a boolean indicating if the high starter battery voltage alarm is active
        """
        return self._data["alarm"]["high_starter_voltage"]

    def get_low_temperature_alarm(self) -> bool:
        """
        Return a boolean indicating if the low temperature alarm is active
        """
        return self._data["alarm"]["low_temperature"]

    def get_high_temperature_alarm(self) -> bool:
        """
        Return a boolean indicating if the high temperature alarm is active
        """
        return self._data["alarm"]["high_temperature"]

    def get_aux_mode(self) -> AuxMode:
        """
        Return an enum indicating the current auxiliary input mode
        """
        return self._data["aux_mode"]

    def get_temperature(self) -> Optional[float]:
        """
        Return the temperature in Celsius if the aux input is set to temperature
        """
        temp = self._data.get("temperature_kelvin")
        if temp:
            return kelvin_to_celsius(temp)
        return None

    def get_starter_voltage(self) -> Optional[float]:
        """
        Return the starter battery voltage in volts if the aux input is set to starter battery
        """
        return self._data.get("starter_voltage")


class DcEnergyMeter(Device):

    PACKET = Struct(
        "meter_type" / Int16sl,
        # Voltage reading in 0.1v increments
        "voltage" / Int16ul,
        "alarm"
        / BitStruct(
            "mid_voltage" / Flag,
            "high_temperature" / Flag,
            "low_temperature" / Flag,
            "high_starter_voltage" / Flag,
            "low_starter_voltage" / Flag,
            "low_soc" / Flag,
            "high_voltage" / Flag,
            "low_voltage" / Flag,
        ),
        # Unknown byte
        "uk_1b" / Int8sl,
        # Value of the auxillary input
        "aux" / Int16ul,
        # The upper 22 bits indicate the current in milliamps
        # The lower 2 bits identify the aux input mode:
        #   0 = Starter battery voltage
        #   1 = Midpoint voltage
        #   2 = Temperature
        #   3 = Disabled
        "current" / Int24sl,
    )

    def parse(self, data: bytes) -> DcEnergyMeterData:
        decrypted = self.decrypt(data)
        pkt = self.PACKET.parse(decrypted)

        aux_mode = AuxMode(pkt.current & 0b11)

        parsed = {
            "meter_type": MeterType(pkt.meter_type),
            "aux_mode": aux_mode,
            "current": (pkt.current >> 2) / 1000,
            "voltage": pkt.voltage / 100,
            "alarm": {
                "low_voltage": pkt.alarm.low_voltage,
                "high_voltage": pkt.alarm.high_voltage,
                "low_starter_voltage": pkt.alarm.low_starter_voltage,
                "high_starter_voltage": pkt.alarm.high_starter_voltage,
                "low_temperature": pkt.alarm.low_temperature,
                "high_temperature": pkt.alarm.high_temperature,
            },
        }

        if aux_mode == AuxMode.STARTER_VOLTAGE:
            # Starter voltage is treated as signed
            parsed["starter_voltage"] = (
                Int16sl.parse((pkt.aux).to_bytes(2, "little")) / 100
            )
        elif aux_mode == AuxMode.TEMPERATURE:
            parsed["temperature_kelvin"] = pkt.aux / 100

        return DcEnergyMeterData(parsed)
